#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2021
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import qlat_utils as q
import copy
import io
import tokenize
import sympy

def get_variables(s):
    """
    `s` is a string
    return a list of variables contained in `s`
    """
    variables = []
    g = tokenize.tokenize(io.BytesIO(s.encode('utf-8')).readline)  # tokenize the string
    for toknum, tokval, _, _, _ in g:
        if toknum == tokenize.NAME:
            variables.append(tokval)
    variables = sorted(list(set(variables)))
    return variables

class Factor:

    """
    self.code
    self.variables
    self.otype
    #
    self.otype in [ "Expr", "Var", ]
    """

    def __init__(self, code, variables=None, otype=None):
        self.code = code
        self.otype = otype
        self.variables = variables
        if self.variables is None:
            self.variables = get_variables(self.code)
        if self.otype is None:
            if code.isidentifier():
                self.otype = "Var"
            else:
                self.otype = "Expr"

    def __repr__(self) -> str:
        return f"ea.Factor({self.code},{self.variables},{self.otype})"

    def __str__(self) -> str:
        return self.compile_py()

    def compile_py(self, var_dict=None) -> str:
        if (var_dict is None) or (self.otype != "Var") or (self.code not in var_dict):
            return f"{self.code}"
        return var_dict[self.code]

### ------

class Term:

    """
    self.coef
    self.factors
    #
    coef should be a number or a sympy expression
    """

    def __init__(self, factors, coef = 1):
        self.coef = coef
        self.factors = factors

    def __repr__(self) -> str:
        return f"ea.Term({self.factors},{self.coef})"

    def __str__(self) -> str:
        return self.compile_py()

    def sort(self) -> None:
        self.factors.sort(key = repr)

    def simplify_coef(self) -> None:
        self.coef = sympy.simplify(self.coef)

    def compile_py(self, var_dict=None) -> str:
        fs = [ f"({f.compile_py(var_dict)})" for f in self.factors ]
        if self.coef == 1:
            if self.factors == []:
                return "1"
            else:
                return '*'.join(fs)
        return '*'.join([ compile_py_complex(self.coef), ] + fs)

### ------

class Expr:

    """
    self.terms
    """

    def __init__(self, terms):
        self.terms = terms

    def __repr__(self) -> str:
        return f"ea.Expr({self.terms})"

    def __str__(self) -> str:
        return self.compile_py()

    def __add__(self, other):
        return Expr(self.terms + mk_expr(other).terms)

    __radd__ = __add__

    def __mul__(self, other):
        other = mk_expr(other)
        terms = []
        for t1 in self.terms:
            for t2 in other.terms:
                coef = t1.coef * t2.coef
                t = Term(t1.factors + t2.factors, coef)
                terms.append(t)
        return Expr(terms)

    def __rmul__(self, other):
        return mk_expr(other) * self

    def __neg__(self):
        return mk_expr(-1) * mk_expr(self)

    def __pos__(self):
        return self

    def __sub__(self, other):
        return mk_expr(self) + mk_expr(-1) * other

    def __rsub__(self, other):
        return mk_expr(other) + mk_expr(-1) * self

    def sort(self) -> None:
        for term in self.terms:
            term.sort()
        self.terms.sort(key = repr)

    def combine_terms(self) -> None:
        self.terms = combine_terms_expr(self).terms

    def drop_zeros(self) -> None:
        self.terms = drop_zero_terms(self).terms

    @q.timer
    def simplify_coef(self) -> None:
        for t in self.terms:
            t.simplify_coef()
        self.drop_zeros()

    @q.timer
    def simplify(self) -> None:
        """
        interface function
        """
        self.sort()
        self.combine_terms()
        self.drop_zeros()

    def is_zero(self) -> bool:
        return not self.terms

    def compile_py(self, var_dict=None) -> str:
        if self.terms:
            return '+'.join([ f"{t.compile_py(var_dict)}" for t in self.terms ])
        else:
            return '0'

### ------

def simplified(x):
    """
    interface function
    """
    x = copy.deepcopy(mk_expr(x))
    x.simplify()
    if not x.terms:
        x = 0
    elif len(x.terms) == 1:
        if not x.terms[0].factors:
            x = x.terms[0].coef
    return x

def coef_simplified(x):
    """
    interface function
    """
    x = copy.deepcopy(mk_expr(x))
    x.simplify_coef()
    return x

def compile_py(x, var_dict=None):
    """
    interface function
    """
    if isinstance(x, (int, float, complex, sympy.Basic,)):
        return compile_py_complex(x)
    else:
        return mk_expr(x).compile_py(var_dict)

def is_zero(x):
    """
    interface function
    """
    if isinstance(x, (int, float, complex, sympy.Basic,)):
        return x == 0
    elif isinstance(x, Expr):
        return x.is_zero()
    else:
        print(x)
        assert False

def mk_expr(x):
    """
    interface function
    """
    if isinstance(x, Expr):
        return x
    elif isinstance(x, Term):
        return Expr([ x, ])
    elif isinstance(x, Factor):
        return Expr([ Term([ x, ]), ])
    elif isinstance(x, (int, float, complex, sympy.Basic,)):
        return Expr([ Term([], x,), ])
    elif isinstance(x, str):
        # str viewed as code segment
        return Expr([ Term([ Factor(x), ]), ])
    else:
        print(x)
        assert False

def compile_py_complex(x):
    """
    interface function
    """
    v = complex(x)
    return f"{v}"

def drop_zero_terms(expr : Expr) -> Expr:
    terms = []
    for t in expr.terms:
        if t.coef != 0:
            terms.append(t)
    return Expr(terms)

def combine_two_terms(t1 : Term, t2 : Term, t1_sig : str, t2_sig : str):
    if t1_sig == t2_sig:
        coef = t1.coef + t2.coef
        if coef == 0:
            return Term([], 0)
        else:
            return Term(t1.factors, coef)
    else:
        return None

def combine_terms_expr(expr : Expr) -> Expr:
    if not expr.terms:
        return expr
    def get_sig(t):
        return f"{t.factors}"
    zero_term = Term([], 0)
    zero_term_sig = get_sig(zero_term)
    signatures = [ get_sig(t) for t in expr.terms ]
    terms = []
    term = expr.terms[0]
    term_sig = signatures[0]
    for t, t_sig in zip(expr.terms[1:], signatures[1:]):
        if term.coef == 0:
            term = t
            term_sig = t_sig
        else:
            ct = combine_two_terms(t, term, t_sig, term_sig)
            if ct is None:
                terms.append(term)
                term = t
                term_sig = t_sig
            elif ct.coef == 0:
                term = zero_term
                term_sig = zero_term_sig
            else:
                term = ct
    if term.coef != 0:
        terms.append(term)
    return Expr(terms)

### ------

def mk_sym(x):
    """
    make a sympy simplified value
    """
    return sympy.simplify(x)

def mk_fac(x):
    """
    make an Expr obj (can be sympy expression)
    """
    return mk_expr(x)

### ------

if __name__ == "__main__":
    a = mk_expr(1)
    b = mk_expr(2)
    c = mk_expr(Factor("a + b"))
    d = mk_expr(Factor("a + b"))
    c = c * b + c + d * d
    print(c)
    c.simplify()
    print(c)
    print(c.compile_py())



import qlat.c as c

from qlat.field import *
from qlat.field_utils import *

class ScalarAction:

    def __init__(self, m_sq, lmbd, alpha):
        self.cdata = c.mk_scalar_action(m_sq, lmbd, alpha)

    def __del__(self):
        assert isinstance(self.cdata, int)
        c.free_scalar_action(self)

    def __imatmul__(self, v1):
        assert isinstance(v1, ScalarAction)
        c.set_scalar_action(self, v1)
        return self

    def m_sq(self):
        return c.get_m_sq_scalar_action(self)

    def lmbd(self):
        return c.get_lmbd_scalar_action(self)

    def lmbd(self):
        return c.get_alpha_scalar_action(self)

    def action_node(self, sf):
        assert isinstance(sf, FieldBase)
        return c.action_node_scalar_action(self, sf)

    def hmc_estimate_mass(self, masses, field_ft, force_ft, phi0):
        assert isinstance(masses, FieldBase)
        assert isinstance(field_ft, FieldBase)
        assert isinstance(force_ft, FieldBase)
        return c.hmc_estimate_mass_scalar_action(self, masses, field_ft, force_ft, phi0)

    def to_mass_factor(self, sin_domega):
        assert isinstance(sin_domega, FieldBase)
        return c.to_mass_factor_scalar_action(self, sin_domega)

    def set_complex_from_double(self, cf, sf):
        assert isinstance(cf, FieldBase)
        assert isinstance(sf, FieldBase)
        return c.set_complex_from_double_scalar_action(self, cf, sf)

    def set_double_from_complex(self, sf, cf):
        assert isinstance(cf, FieldBase)
        assert isinstance(sf, FieldBase)
        return c.set_double_from_complex_scalar_action(self, sf, cf)

    def sum_sq(self, sf):
        assert isinstance(sf, FieldBase)
        return c.sum_sq_scalar_action(self, sf)

    def hmc_m_hamilton_node(self, sf, masses):
        assert isinstance(sf, FieldBase)
        assert isinstance(masses, FieldBase)
        return c.hmc_m_hamilton_node_scalar_action(self, sf, masses)

    def hmc_set_force(self, sm_force, sf):
        assert isinstance(sm_force, FieldBase)
        assert isinstance(sf, FieldBase)
        return c.hmc_set_force_scalar_action(self, sm_force, sf)

    def hmc_field_evolve(self, sf_ft, sm_ft, masses, step_size):
        assert isinstance(sf_ft, FieldBase)
        assert isinstance(sm_ft, FieldBase)
        assert isinstance(masses, FieldBase)
        c.hmc_field_evolve_scalar_action(self, sf_ft, sm_ft, masses, step_size)

    def axial_current_node(self, axial_current, sf):
        assert isinstance(sf, FieldBase)
        assert isinstance(axial_current, FieldBase)
        return c.axial_current_node_scalar_action(self, axial_current, sf)

    def hmc_set_rand_momentum(self, sm_complex, masses, rs):
        assert isinstance(sm_complex, FieldBase)
        assert isinstance(masses, FieldBase)
        assert isinstance(rs, RngState)
        return c.hmc_set_rand_momentum_scalar_action(self, sm_complex, masses, rs)

    def hmc_predict_field(self, field_ft, momentum_ft, masses, vev_sigma):
        assert isinstance(field_ft, FieldBase)
        assert isinstance(momentum_ft, FieldBase)
        assert isinstance(masses, FieldBase)
        return c.hmc_predict_field_scalar_action(self, field_ft, momentum_ft, masses, vev_sigma)

    def get_polar_field(self, polar_field, field):
        assert isinstance(polar_field, FieldBase)
        assert isinstance(field, FieldBase)
        return c.get_polar_field_scalar_action(self, polar_field, field)



from qlat import *

import qlat.c as c

class FlowInfo:

    def __init__(self):
        self.cdata = c.mk_flow_info()

    def __del__(self):
        c.free_flow_info(self)

    def add_flow(self, eo, mu, epsilon, flow_size = 1):
        c.add_flow_flow_info(self, eo, mu, epsilon, flow_size)

    def add_rand_order_flow(self, rng, epsilon, *args):
        if len(args) == 0:
            c.add_rand_order_flow_flow_info(self, rng, epsilon)
        elif len(args) == 1:
            epsilon2 = args[0]
            c.add_rand_order_flow2_flow_info(self, rng, epsilon, epsilon2)
        else:
            raise Exception("add_rand_order_flow")

    def show(self):
        return c.show_flow_info(self)

def gf_flow(gf, gf0, fi):
    assert isinstance(gf, GaugeField)
    assert isinstance(gf0, GaugeField)
    assert isinstance(fi, FlowInfo)
    c.gf_flow(gf, gf0, fi)

def gf_flow_inv(gf, gf1, fi):
    assert isinstance(gf, GaugeField)
    assert isinstance(gf1, GaugeField)
    assert isinstance(fi, FlowInfo)
    c.gf_flow_inv(gf, gf1, fi)

def gf_hamilton_flowed_node(gf0, ga, fi):
    assert isinstance(gf0, GaugeField)
    assert isinstance(ga, GaugeAction)
    assert isinstance(fi, FlowInfo)
    return c.gf_hamilton_flowed_node(gf0, ga, fi)

def set_gm_force_flowed(gm_force, gf0, ga, fi):
    assert isinstance(gm_force, GaugeMomentum)
    assert isinstance(gf0, GaugeField)
    assert isinstance(ga, GaugeAction)
    assert isinstance(fi, FlowInfo)
    c.set_gm_force_flowed(gm_force, gf0, ga, fi)

def set_gm_force_flowed_no_det(gm_force, gm_force_pre, gf0, fi):
    assert isinstance(gm_force, GaugeMomentum)
    assert isinstance(gm_force_pre, GaugeMomentum)
    assert isinstance(gf0, GaugeField)
    assert isinstance(fi, FlowInfo)
    c.set_gm_force_flowed_no_det(gm_force, gm_force_pre, gf0, fi)



from qlat_utils import *
import qlat.c as c

from qlat.propagator import *

@timer
def contract_pion_field(prop, tslice):
    ld = LatData()
    if isinstance(prop, Prop):
        c.contract_pion_field(ld, prop, tslice)
    elif isinstance(prop, SelProp):
        c.contract_pion_sfield(ld, prop, tslice)
    else:
        raise Exception("contract_pion_field")
    return ld



from qlat_utils import *
import qlat.c as c

from qlat.field import *
from qlat.qcd import *
from qlat.gauge_action import *
from qlat.elem_type import *

import math

class GaugeMomentum(FieldColorMatrix):

    def __init__(self, geo = None):
        super().__init__(geo, 4)

    def set_rand(self, rng, sigma = 1.0):
        set_rand_gauge_momentum(self, sigma, rng)

def set_rand_gauge_momentum(gm, sigma, rng):
    assert isinstance(gm, GaugeMomentum)
    assert isinstance(sigma, float)
    assert isinstance(rng, RngState)
    return c.set_rand_gauge_momentum(gm, sigma, rng)

def gm_hamilton_node(gm):
    assert isinstance(gm, GaugeMomentum)
    return c.gm_hamilton_node(gm)

def gf_hamilton_node(gf, ga):
    assert isinstance(gf, GaugeField)
    assert isinstance(ga, GaugeAction)
    return c.gf_hamilton_node(gf, ga)

def gf_evolve(gf, gm, step_size):
    assert isinstance(gm, GaugeMomentum)
    return c.gf_evolve(gf, gm, step_size)

def set_gm_force(gm_force, gf, ga):
    assert isinstance(gm_force, GaugeMomentum)
    assert isinstance(gf, GaugeField)
    assert isinstance(ga, GaugeAction)
    return c.set_gm_force(gm_force, gf, ga)

@timer_verbose
def metropolis_accept(delta_h, traj, rs):
    flag_d = 0.0
    accept_prob = 0.0
    if get_id_node() == 0:
        if delta_h <= 0.0:
            accept_prob = 1.0
            flag_d = 1.0
        else:
            accept_prob = math.exp(-delta_h)
            rand_num = rs.u_rand_gen(1.0, 0.0)
            if rand_num <= accept_prob:
                flag_d = 1.0
    flag_d = glb_sum(flag_d)
    accept_prob = glb_sum(accept_prob)
    flag = flag_d > 0.5
    displayln_info(f"metropolis_accept: flag={flag:d} with accept_prob={accept_prob * 100.0:.1f}% delta_h={delta_h:.16f} traj={traj}")
    return flag, accept_prob

@timer
def gm_evolve_fg_pure_gauge(gm, gf_init, ga, fg_dt, dt):
    geo = gf_init.geo()
    gf = GaugeField(geo)
    gf @= gf_init
    gm_force = GaugeMomentum(geo)
    set_gm_force(gm_force, gf, ga)
    gf_evolve(gf, gm_force, fg_dt)
    set_gm_force(gm_force, gf, ga)
    gm_force *= dt
    gm += gm_force

@timer_verbose
def run_hmc_evolve_pure_gauge(gm, gf, ga, rs, n_step, md_time = 1.0):
    energy = gm_hamilton_node(gm) + gf_hamilton_node(gf, ga)
    dt = md_time / n_step
    lam = 0.5 * (1.0 - 1.0 / math.sqrt(3.0));
    theta = (2.0 - math.sqrt(3.0)) / 48.0;
    ttheta = theta * dt * dt * dt;
    gf_evolve(gf, gm, lam * dt)
    for i in range(n_step):
        gm_evolve_fg_pure_gauge(gm, gf, ga, 4.0 * ttheta / dt, 0.5 * dt);
        gf_evolve(gf, gm, (1.0 - 2.0 * lam) * dt);
        gm_evolve_fg_pure_gauge(gm, gf, ga, 4.0 * ttheta / dt, 0.5 * dt);
        if i < n_step - 1:
            gf_evolve(gf, gm, 2.0 * lam * dt);
        else:
            gf_evolve(gf, gm, lam * dt);
    gf.unitarize()
    delta_h = gm_hamilton_node(gm) + gf_hamilton_node(gf, ga) - energy;
    delta_h = glb_sum(delta_h)
    return delta_h

@timer_verbose
def run_hmc_pure_gauge(gf, ga, traj, rs, *, is_reverse_test = False, n_step = 6, md_time = 1.0, is_always_accept = False):
    fname = get_fname()
    rs = rs.split(f"{traj}")
    geo = gf.geo()
    gf0 = GaugeField(geo)
    gf0 @= gf
    gm = GaugeMomentum(geo)
    gm.set_rand(rs.split("set_rand_gauge_momentum"), 1.0)
    delta_h = run_hmc_evolve_pure_gauge(gm, gf0, ga, rs, n_step, md_time)
    if is_reverse_test:
        gm_r = GaugeMomentum(geo)
        gm_r @= gm
        gf0_r = GaugeField(geo)
        gf0_r @= gf0
        delta_h_rev = run_hmc_evolve_pure_gauge(gm_r, gf0_r, ga, rs, n_step, -md_time)
        gf0_r -= gf;
        displayln_info(f"{fname}: reversed delta_diff: {delta_h + delta_h_rev} / {delta_h}")
        displayln_info(f"{fname}: reversed gf_diff: {qnorm(gf0_r)} / {qnorm(gf0)}")
    flag, accept_prob = metropolis_accept(delta_h, traj, rs.split("metropolis_accept"))
    if flag or is_always_accept:
        displayln_info(f"{fname}: update gf (traj={traj})")
        gf @= gf0



from .. import rbc_ukqcd_params as rup

# Christoph ensemble 2, ml = 0.0055, ms = 0.0368, Ls = 32, beta = 2.13

job_tag = "24IH02"

dict_params = {}

rup.dict_params[job_tag] = dict_params

dict_params["job_tag"] = job_tag

dict_params["total_site"] = [ 24, 24, 24, 48, ]

dict_params["load_config_params"] = { "twist_boundary_at_boundary":[ 0.0, 0.0, 0.0, -0.5, ] }

def mk_fermion_params(inv_type, inv_acc):
    params = {}
    params["M5"] = 1.8
    params["boundary_phases"] = [ 1.0, 1.0, 1.0, 1.0, ] # twist boundary after loading configuration
    params["b"] = 1.5
    params["c"] = 0.5
    if inv_type == 0:
        params["mass"] = 0.0055
    elif inv_type == 1:
        params["mass"] = 0.0368
    elif inv_type == 2:
        params["mass"] = 0.2
    else:
        assert False
    if inv_acc == 0 or inv_acc == 1 or inv_acc == 2:
        params["Ls"] = 32
    else:
        assert False
    return params

def mk_dict_fermion_params():
    params = {}
    for inv_type in [ 0, 1, 2, ]:
        params[inv_type] = {}
        for inv_acc in [ 0, 1, 2, ]:
            params[inv_type][inv_acc] = mk_fermion_params(inv_type, inv_acc)
    return params

dict_params["fermion_params"] = mk_dict_fermion_params()



from qlat.field import *

def set_checkers(field):
    # no longer needed?
    assert field.ctype == c.ElemTypeDouble
    c.set_checkers_double_field(field)

def set_double_from_complex(field, cf):
    assert isinstance(cf, FieldBase)
    assert field.ctype == c.ElemTypeDouble
    assert cf.ctype == c.ElemTypeComplex
    c.set_double_from_complex_field(field,cf)

def set_complex_from_double(field, sf):
    assert isinstance(sf, FieldBase)
    assert field.ctype == c.ElemTypeComplex
    assert sf.ctype == c.ElemTypeDouble
    c.set_complex_from_double_field(field,sf)

def set_abs_from_complex(field, cf):
    assert isinstance(cf, FieldBase)
    assert field.ctype == c.ElemTypeDouble
    assert cf.ctype == c.ElemTypeComplex
    c.set_abs_from_complex_field(field,cf)

def set_ratio_double(field, sf1, sf2):
    assert isinstance(sf1, FieldBase)
    assert isinstance(sf2, FieldBase)
    assert field.ctype == c.ElemTypeDouble
    assert sf1.ctype == c.ElemTypeDouble
    assert sf2.ctype == c.ElemTypeDouble
    c.set_ratio_double_field(field,sf1,sf2)

def less_than_double(field, sf2, mask):
    assert isinstance(sf2, FieldBase)
    assert isinstance(mask, FieldBase)
    assert field.ctype == c.ElemTypeDouble
    assert sf2.ctype == c.ElemTypeDouble
    assert mask.ctype == c.ElemTypeDouble
    c.less_than_double_field(field,sf2,mask)

def invert_double(field):
    assert field.ctype == c.ElemTypeDouble
    c.invert_double_field(field)

def multiply_double(field, factor):
    assert isinstance(field, FieldBase)
    assert isinstance(factor, FieldBase)
    assert factor.ctype is c.ElemTypeDouble
    c.multiply_double_field(field, factor)



import qlat as q
from . import rbc_ukqcd_params as rup
from .rbc_ukqcd_params import set_param, get_param

import numpy as np

import pprint
import functools
import os

save_path_default = "results"

load_path_list = [ "results", ]

def get_save_path(fn):
    return os.path.join(save_path_default, fn)

def get_load_path(*fns):
    def get(fn):
        if fn is None:
            return None
        elif isinstance(fn, (tuple, list)):
            for f in fn:
                p = get(f)
                if p is not None:
                    return p
        else:
            for path in load_path_list:
                p = os.path.join(path, fn)
                if q.does_file_exist_qar_sync_node(p):
                    return p
        return None
    return get(fns)

# ----------

@q.timer_verbose
def check_job(job_tag, traj, fns_produce, fns_need):
    """
    return False if config is finished or unavailable
    """
    is_job_done = True
    for fn in fns_produce:
        if get_load_path(fn) is None:
            q.displayln_info(f"check_job: {job_tag} {traj} to do as '{fn}' does not exist.")
            is_job_done = False
            break
    if is_job_done:
        return False
    #
    is_job_avail = True
    for fn in fns_need:
        if get_load_path(fn) is None:
            q.displayln_info(f"check_job: {job_tag} {traj} unavailable as '{fn}' does not exist.")
            is_job_avail = False
            break
    if not is_job_avail:
        return False
    #
    q.check_stop()
    q.check_time_limit()
    #
    assert not is_job_done and is_job_avail
    #
    return True

# ----------

@q.timer_verbose
def run_gf(job_tag, traj):
    path_gf = get_load_path(
            f"{job_tag}/configs/ckpoint_lat.{traj}",
            f"{job_tag}/configs/ckpoint_lat.IEEE64BIG.{traj}",
            )
    if path_gf is None:
        if job_tag[:5] == "test-":
            if not q.obtain_lock(f"locks/{job_tag}-{traj}-gauge_field"):
                return None
            total_site = rup.get_total_site(job_tag)
            gf = rup.mk_sample_gauge_field_v3(job_tag, f"{traj}")
            path_gf = get_save_path(f"{job_tag}/configs/ckpoint_lat.{traj}")
            gf.save(path_gf)
            q.release_lock()
        else:
            @q.timer_verbose
            def load_gf():
                assert False
            return load_gf
    get_gf = rup.load_config_lazy(job_tag, path_gf)
    return get_gf

@q.timer_verbose
def run_gt(job_tag, traj, get_gf):
    if None in [ get_gf, ]:
        return None
    tfn = f"{job_tag}/gauge-transform/traj-{traj}.field"
    path_gt = get_load_path(tfn)
    if path_gt is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-gauge_fix_coulomb"):
            gf = get_gf()
            import qlat_gpt as qg
            gt = qg.gauge_fix_coulomb(gf)
            gt.save_cps(get_save_path(f"{job_tag}/gauge-transform/traj-{traj}.gfix"))
            gt.save_double(get_save_path(tfn))
            q.release_lock()
            return lambda : gt
        else:
            return None
    @q.timer_verbose
    def load_gt():
        path_gt = get_load_path(tfn)
        assert path_gt is not None
        gt = q.GaugeTransform()
        gt.load_double(path_gt)
        # ADJUST ME
        # import qlat_gpt as qg
        # qg.check_gauge_fix_coulomb(get_gf(), gt)
        #
        return gt
    get_gt = q.lazy_call(load_gt)
    return get_gt

# ----------

@q.timer
def mk_rand_wall_src_info_n_exact(job_tag, traj, inv_type):
    params = rup.dict_params[job_tag]
    n_exact = params["n_exact_wsrc"]
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_wall_src_info")
    inv_acc_s = 1
    inv_acc_e = 2
    total_site = rup.get_total_site(job_tag)
    t_size = total_site[3]
    wi_s = [ [ t, inv_type, inv_acc_s, ] for t in range(t_size) ]
    mask = [ False, ] * t_size
    for i in range(n_exact):
        t_e = rs.rand_gen() % t_size
        mask[t_e] = True
    wi_e = []
    for t in range(t_size):
        if mask[t]:
            wi_e.append([ t, inv_type, inv_acc_e, ])
    wi = wi_e + wi_s
    for i in range(len(wi)):
        wi[i] = [ i, ] + wi[i]
    return wi

@q.timer
def mk_rand_wall_src_info_prob(job_tag, traj, inv_type):
    params = rup.dict_params[job_tag]
    prob = params["prob_exact_wsrc"]
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_wall_src_info_prob")
    inv_acc_s = 1
    inv_acc_e = 2
    total_site = rup.get_total_site(job_tag)
    t_size = total_site[3]
    wi_s = [ [ t, inv_type, inv_acc_s, ] for t in range(t_size) ]
    wi_e = []
    for t in range(t_size):
        if rs.u_rand_gen() < prob:
            wi_e.append([ t, inv_type, inv_acc_e, ])
    wi = wi_e + wi_s
    for i in range(len(wi)):
        wi[i] = [ i, ] + wi[i]
    return wi

def get_prob_exact_wsrc(job_tag):
    params = rup.dict_params[job_tag]
    if "prob_exact_wsrc" in params:
        return params["prob_exact_wsrc"]
    n_exact = params["n_exact_wsrc"]
    total_site = rup.get_total_site(job_tag)
    return 1 - (1 - 1 / total_site[3])**n_exact

@q.timer
def mk_rand_wall_src_info(job_tag, traj, inv_type):
    """
    wi is a list of [ idx tslice inv_type inv_acc ]
    """
    params = rup.dict_params[job_tag]
    if "prob_exact_wsrc" not in params:
        return mk_rand_wall_src_info_n_exact(job_tag, traj, inv_type)
    return mk_rand_wall_src_info_prob(job_tag, traj, inv_type)

@q.timer
def save_wall_src_info(wi, path):
    """
    wi is a list of  [ idx tslice inv_type inv_acc ]
    """
    if 0 != q.get_id_node():
        return None
    lines = [ " ".join([ f"{v:5d}" for v in l ]) for l in wi ]
    content = "\n".join(lines + [ "", ])
    q.qtouch(path, content)

@q.timer
def load_wall_src_info(path):
    assert path is not None
    """
    wi is a list of [ idx tslice inv_type inv_acc ]
    """
    dt = q.qload_datatable_sync_node(path, True)
    t = [ list(map(int, l)) for l in dt ]
    wi = [ [ l[0], l[1], l[2], l[3], ] for l in t ]
    return wi

@q.timer_verbose
def run_wi(job_tag, traj):
    tfn_l = f"{job_tag}/wall-src-info-light/traj-{traj}.txt"
    tfn_s = f"{job_tag}/wall-src-info-strange/traj-{traj}.txt"
    path_light = get_load_path(tfn_l)
    if path_light is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-wi"):
            wi_light = mk_rand_wall_src_info(job_tag, traj, inv_type = 0)
            save_wall_src_info(wi_light, get_save_path(tfn_l));
            q.release_lock()
        else:
            return None
    path_strange = get_load_path(tfn_s)
    if path_strange is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-wi"):
            wi_strange = mk_rand_wall_src_info(job_tag, traj, inv_type = 1)
            save_wall_src_info(wi_strange, get_save_path(tfn_s));
            q.release_lock()
        else:
            return None
    @q.timer_verbose
    def load():
        wi_light = load_wall_src_info(get_load_path(tfn_l))
        wi_strange = load_wall_src_info(get_load_path(tfn_s))
        return wi_light + wi_strange
    return q.lazy_call(load)

# ----------

def get_n_points_psel(job_tag):
    assert job_tag in rup.dict_params
    assert "n_points_psel" in rup.dict_params[job_tag]
    return get_param(job_tag, "n_points_psel")

@q.timer
def mk_rand_psel(job_tag, traj):
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_psel")
    total_site = rup.get_total_site(job_tag)
    n_points = get_n_points_psel(job_tag)
    psel = q.PointsSelection()
    psel.set_rand(rs, total_site, n_points)
    return psel

@q.timer_verbose
def run_psel(job_tag, traj):
    tfn = f"{job_tag}/point-selection/traj-{traj}.txt"
    path_psel = get_load_path(tfn)
    if path_psel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-psel"):
            psel = mk_rand_psel(job_tag, traj)
            psel.save(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    #
    @q.timer_verbose
    def load_psel():
        path_psel = get_load_path(tfn)
        assert path_psel is not None
        total_site = rup.get_total_site(job_tag)
        psel = q.PointsSelection()
        psel.load(path_psel, q.Geometry(total_site))
        assert psel.n_points() == get_n_points_psel(job_tag)
        return psel
    return q.lazy_call(load_psel)

# ----------

def get_n_points_pi(job_tag, traj, inv_type, inv_acc):
    assert job_tag in rup.dict_params
    assert "n_points" in rup.dict_params[job_tag]
    return rup.dict_params[job_tag]["n_points"][inv_type][inv_acc]

@q.timer
def mk_rand_point_src_info(job_tag, traj, psel):
    # pi is a list of [ idx xg inv_type inv_acc ]
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_point_src_info")
    xg_list = psel.to_list()
    assert len(xg_list) == get_n_points_pi(job_tag, traj, 0, 0)
    g_pi = [ [] for _ in xg_list ]
    for inv_type in [ 0, 1, ]:
        for inv_acc in [ 0, 1, 2, ]:
            for i in range(get_n_points_pi(job_tag, traj, inv_type, inv_acc)):
                g_pi[i].append([ xg_list[i], inv_type, inv_acc ])
    pi = []
    for g in g_pi:
        pi += g
    for i in range(len(pi)):
        pi[i] = [ i, ] + pi[i]
    return pi

@q.timer
def save_point_src_info(pi, path):
    # pi is a list of [ idx xg inv_type inv_acc ]
    if 0 != q.get_id_node():
        return None
    def mk_line(l):
        [ idx, xg, inv_type, inv_acc ] = l
        return f"{idx:5d}    {xg[0]:3d} {xg[1]:3d} {xg[2]:3d} {xg[3]:3d}    {inv_type:3d} {inv_acc:3d}"
    lines = list(map(mk_line, pi))
    content = "\n".join([ f"{len(lines)}" ] + lines + [ "" ])
    q.qtouch(path, content)

@q.timer
def load_point_src_info(path):
    # pi is a list of [ idx xg inv_type inv_acc ]
    dt = q.qload_datatable_sync_node(path, True)
    t = [ list(map(int, l)) for l in dt ][1:]
    pi = [ [ l[0], l[1:5], l[5], l[6], ] for l in t ]
    return pi

@q.timer_verbose
def run_pi(job_tag, traj, get_psel):
    tfn = f"{job_tag}/point-src-info/traj-{traj}.txt"
    path = get_load_path(tfn)
    if path is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-pi"):
            pi = mk_rand_point_src_info(job_tag, traj, get_psel())
            save_point_src_info(pi, get_save_path(tfn));
            q.release_lock()
        else:
            return None
    @q.timer_verbose
    def load():
        path = get_load_path(tfn)
        assert path is not None
        pi = load_point_src_info(path)
        return pi
    return q.lazy_call(load)

# ----------

@q.timer
def load_point_distribution(job_tag):
    """
    return point_distribution
    where
    point_distribution[xg_rel] = probability of the relative coordinate of a point relative to a selected point equals to ``xg_rel``.
    xg_rel = (x, y, z, t,)
    x >= y >= z >= 0 and t >= 0
    n_points = get_n_points_psel(job_tag)
    """
    n_points = get_n_points_psel(job_tag)
    tfn = f"{job_tag}/point-distribution/point-distribution.txt"
    path = get_load_path(tfn)
    if path is None:
        return None
    dt = q.qload_datatable_sync_node(path, True)
    point_distribution = dict()
    for l in dt:
        x, y, z, t, prob = l
        x = int(x)
        y = int(y)
        z = int(z)
        t = int(t)
        point_distribution[(x, y, z, t,)] = prob * (n_points - 1) / n_points
    point_distribution[(0, 0, 0, 0,)] = 1.0 / n_points
    return point_distribution

def classify_rel_coordinate(xg_rel_arrary, total_site_array):
    """
    xg_rel_arrary = np.array(xg_rel)
    total_site_array = np.array(total_site)
    """
    total_site_half = total_site_array // 2
    xg_rel_arrary = xg_rel_arrary % total_site_array
    xg_rel_abs = total_site_half - abs(xg_rel_arrary - total_site_half)
    x, y, z, t = xg_rel_abs
    x, y, z = sorted([x, y, z])
    return (x, y, z, t,)

def get_point_xrel_prob(xg_rel_arrary, total_site_array, point_distribution, n_points):
    """
    xg_rel_arrary = np.array(xg_rel)
    total_site_array = np.array(total_site)
    point_distribution = load_point_distribution(job_tag)
    n_points = get_n_points_psel(job_tag)
    """
    if point_distribution is None:
        if np.all(xg_rel_arrary == 0):
            return 1.0 / n_points
        else:
            total_volume = np.prod(total_site_array)
            return (n_points - 1) / n_points / (total_volume - 1)
    xg_rel = classify_rel_coordinate(xg_rel_arrary, total_site_array)
    prob = point_distribution[xg_rel]
    return prob

# ----------

@q.timer
def mk_rand_fsel(job_tag, traj, n_per_tslice):
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_fsel")
    total_site = rup.get_total_site(job_tag)
    fsel = q.FieldSelection()
    fsel.set_rand(rs, total_site, n_per_tslice)
    return fsel

@q.timer
def mk_fselc(fsel, psel):
    fselc = fsel.copy()
    fselc.add_psel(psel)
    return fselc

@q.timer_verbose
def run_fsel(job_tag, traj, get_psel):
    if get_psel is None:
        return None
    tfn = f"{job_tag}/field-selection/traj-{traj}.field"
    path_fsel = get_load_path(tfn)
    total_site = rup.get_total_site(job_tag)
    n_per_tslice = total_site[0] * total_site[1] * total_site[2] // 16
    if path_fsel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-fsel"):
            fsel = mk_rand_fsel(job_tag, traj, n_per_tslice)
            fsel.save(get_save_path(tfn))
            fselc = mk_fselc(fsel, get_psel())
            q.release_lock()
            return lambda : ( fsel, fselc, )
        else:
            return None
    @q.timer_verbose
    def load_fsel():
        path_fsel = get_load_path(tfn)
        assert path_fsel is not None
        fsel = q.FieldSelection()
        total_size = fsel.load(path_fsel, n_per_tslice)
        assert total_size > 0
        fselc = mk_fselc(fsel, get_psel())
        return fsel, fselc
    return q.lazy_call(load_fsel)

# ----------

@q.timer
def mk_rand_fsel_smear(job_tag, traj, n_per_tslice_smear):
    rs = q.RngState(f"seed {job_tag} {traj}").split("mk_rand_fsel_smear")
    total_site = rup.get_total_site(job_tag)
    fsel = q.FieldSelection()
    fsel.set_rand(rs, total_site, n_per_tslice_smear)
    return fsel

@q.timer_verbose
def run_psel_smear(job_tag, traj):
    # return lambda : psel_smear
    # psel_smear should randomly select same number of point on each tslice
    tfn = f"{job_tag}/point-selection-smear/traj-{traj}.txt"
    path_psel = get_load_path(tfn)
    total_site = rup.get_total_site(job_tag)
    if path_psel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-psel-smear"):
            n_per_tslice_smear = rup.dict_params[job_tag]["n_per_tslice_smear"]
            fsel = mk_rand_fsel_smear(job_tag, traj, n_per_tslice_smear)
            psel = fsel.to_psel()
            psel.save(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    #
    @q.timer_verbose
    def load_psel():
        path_psel = get_load_path(tfn)
        assert path_psel is not None
        total_site = rup.get_total_site(job_tag)
        psel = q.PointsSelection()
        psel.load(path_psel, q.Geometry(total_site))
        return psel
    return q.lazy_call(load_psel)

# ----------

@q.timer
def run_gf_ape(job_tag, get_gf):
    if get_gf is None:
        return None
    coef = rup.dict_params[job_tag]["gf_ape_smear_coef"]
    step = rup.dict_params[job_tag]["gf_ape_smear_step"]
    #
    @q.timer_verbose
    def run():
        gf = get_gf()
        gf_ape = gf.copy()
        q.gf_spatial_ape_smear(gf_ape, coef, step)
        gf_ape = q.mk_left_expanded_gauge_field(gf_ape)
        return gf_ape
    return q.lazy_call(run)

# ----------

@q.timer
def compute_eig(gf, job_tag, inv_type = 0, inv_acc = 0, *, path = None):
    """
    return a function ``get_eig''
    ``get_eig()'' return the ``eig''
    """
    from . import rbc_ukqcd as ru
    load_eig = ru.load_eig_lazy(get_load_path(path), job_tag)
    if load_eig is not None:
        return load_eig
    # evec, evals = ru.mk_eig(gf, job_tag, inv_type, inv_acc)
    basis, cevec, smoothed_evals = ru.mk_ceig(gf, job_tag, inv_type, inv_acc)
    eig = [ basis, cevec, smoothed_evals, ]
    ru.save_ceig(get_save_path(path + ".partial"), eig, job_tag, inv_type, inv_acc);
    q.qrename_info(get_save_path(path + ".partial"), get_save_path(path))
    test_eig(gf, eig, job_tag, inv_type)
    def get_eig():
        return eig
    return get_eig

@q.timer
def test_eig(gf, eig, job_tag, inv_type):
    from . import rbc_ukqcd as ru
    geo = gf.geo()
    src = q.FermionField4d(geo)
    src.set_rand(q.RngState("test_eig:src.set_rand"))
    q.displayln_info(f"src norm {src.qnorm():.10E}")
    sol_ref = ru.get_inv(gf, job_tag, inv_type, inv_acc = 2, eig = eig, eps = 1e-10, mpi_split = False, qtimer = False) * src
    q.displayln_info(f"sol_ref norm {sol_ref.qnorm():.10E} with eig")
    for inv_acc in [ 0, 1, 2, ]:
        sol = ru.get_inv(gf, job_tag, inv_type, inv_acc, eig = eig, mpi_split = False, qtimer = False) * src
        sol -= sol_ref
        q.displayln_info(f"sol diff norm {sol.qnorm()} inv_acc={inv_acc} with eig")
        if inv_acc in [ 0, 1, ]:
            sol = ru.get_inv(gf, job_tag, inv_type, inv_acc, mpi_split = False, qtimer = False) * src
            sol -= sol_ref
            q.displayln_info(f"sol diff norm {sol.qnorm()} inv_acc={inv_acc} without eig")

@q.timer_verbose
def run_eig(job_tag, traj, get_gf):
    if None in [ get_gf, ]:
        return None
    from . import rbc_ukqcd as ru
    get_eig = ru.load_eig_lazy(get_load_path(f"{job_tag}/eig/traj-{traj}"), job_tag)
    if get_eig is None and get_gf is not None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-run-eig"):
            get_eig = compute_eig(get_gf(), job_tag, inv_type = 0, path = f"{job_tag}/eig/traj-{traj}")
            q.release_lock()
            return get_eig
        else:
            return None
    else:
        return get_eig

@q.timer_verbose
def run_eig_strange(job_tag, traj, get_gf):
    # if failed, return None
    # if no parameter, return lambda : None
    if None in [ get_gf, ]:
        return None
    if 1 not in rup.dict_params[job_tag]["clanc_params"]:
        fn = f"{job_tag}/eig-strange/traj-{traj}/no-eig-parameters.txt"
        if get_load_path(fn) is None:
            q.qtouch_info(get_save_path(fn))
        return lambda : None
    from . import rbc_ukqcd as ru
    get_eig = ru.load_eig_lazy(get_load_path(f"{job_tag}/eig-strange/traj-{traj}"), job_tag)
    if get_eig is None and get_gf is not None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-run-eig-strange"):
            get_eig = compute_eig(get_gf(), job_tag, inv_type = 1, path = f"{job_tag}/eig-strange/traj-{traj}")
            q.release_lock()
            return get_eig
        else:
            return None
    else:
        return get_eig

# ----------

@functools.lru_cache(maxsize=None)
def get_r_list(job_tag):
    total_site = q.Coordinate(rup.dict_params[job_tag]["total_site"])
    r_limit = q.get_r_limit(total_site)
    r_list = q.mk_r_list(r_limit, r_all_limit=28.0, r_scaling_factor=5.0)
    # r_list = q.mk_r_list(r_limit, r_all_limit=0.0, r_scaling_factor=5.0) # old choice
    return r_list

@functools.lru_cache(maxsize=None)
def get_r_sq_interp_idx_coef_list(job_tag):
    """
    Return [ (r_idx_low, r_idx_high, coef_low, coef_high,), ... ] indexed by r_sq
    """
    r_list = get_r_list(job_tag)
    return q.mk_r_sq_interp_idx_coef_list(r_list)

@q.timer_verbose
def run_r_list(job_tag):
    fn = f"{job_tag}/r_list/r_list.lat"
    r_list = get_r_list(job_tag)
    ld = q.mk_lat_data([
        [ "r_idx", len(r_list), ],
        ])
    ld.from_numpy(np.array(r_list))
    if get_load_path(fn) is not None:
        ld_load = q.load_lat_data(get_load_path(fn))
        assert ld.is_match(ld_load)
        ld_diff = ld - ld_load
        assert ld_diff.qnorm() < 1e-20
        return
    ld.save(get_save_path(fn))

# ----------



from .. import rbc_ukqcd_params as rup

def mk_test_l_t_list():
    lt_list = []
    for l in [ 2, 4, 8, 12, 16, 24, 32, 48, 64, 96, 128, ]:
        for t in [ 4, 8, 12, 16, 24, 32, 48, 64, 96, 128, 192, 256, ]:
            lt_list.append([l, t,])
    return lt_list

def mk_fermion_params(inv_type, inv_acc):
    params = {}
    params["M5"] = 1.8
    params["boundary_phases"] = [ 1.0, 1.0, 1.0, 1.0, ] # twist boundary after loading configuration
    if inv_acc in [ 0, 1, 2, ]:
        params["b"] = 1.5
        params["c"] = 0.5
        params["Ls"] = 8
    else:
        assert False
    if inv_type == 0:
        params["mass"] = 0.01
    elif inv_type == 1:
        params["mass"] = 0.04
    elif inv_type == 2:
        params["mass"] = 0.20
    else:
        assert False
    return params

def mk_dict_fermion_params():
    params = {}
    for inv_type in [ 0, 1, 2, ]:
        params[inv_type] = {}
        for inv_acc in [ 0, 1, 2, ]:
            params[inv_type][inv_acc] = mk_fermion_params(inv_type, inv_acc)
    return params

def mk_lanc_params(job_tag, inv_type, inv_acc):
    assert inv_type in [ 0, 1, ]
    assert inv_acc == 0
    fermion_params = mk_dict_fermion_params()[inv_type][inv_acc]
    pit_params = { "eps": 0.01, "maxiter": 500, "real": True }
    if job_tag == "test-4nt8":
        cheby_params = {"low": 0.10, "high": 5.5, "order": 50}
    elif job_tag == "test-4nt16":
        cheby_params = {"low": 0.05, "high": 5.5, "order": 50}
    else:
        cheby_params = {"low": 0.05, "high": 5.5, "order": 50}
    irl_params = {
            "Nstop": 20,
            "Nk": 25,
            "Nm": 30,
            "resid": 1e-8,
            "betastp": 0.0,
            "maxiter": 20,
            "Nminres": 1,
            # "maxapply": 100
            }
    return {
            "fermion_params": fermion_params,
            "pit_params": pit_params,
            "cheby_params": cheby_params,
            "irl_params": irl_params,
            }

def mk_clanc_params(job_tag, inv_type, inv_acc):
    assert inv_type in [ 0, 1, ]
    assert inv_acc == 0
    block = [ 2, 2, 2, 2, ]
    nbasis = 20
    if inv_type == 0:
        cheby_params = { "low": 0.20, "high": 5.5, "order": 50, }
    elif inv_type == 1:
        cheby_params = { "low": 0.30, "high": 5.5, "order": 50, }
    else:
        assert False
    irl_params = {
            "Nstop": 30,
            "Nk": 35,
            "Nm": 40,
            "resid": 1e-8,
            "betastp": 0.0,
            "maxiter": 20,
            "Nminres": 1,
            # "maxapply": 100
            }
    smoother_params = { "eps": 1e-8, "maxiter": 10, }
    save_params = { "nsingle": 10, "mpi": [ 1, 1, 1, 4, ], }
    return {
            "block": block,
            "nbasis": nbasis,
            "cheby_params": cheby_params,
            "irl_params": irl_params,
            "smoother_params": smoother_params,
            "save_params": save_params,
            }

def setup_params():
    for l, t in mk_test_l_t_list():
        job_tag = f"test-{l}nt{t}"
        dict_params = {}
        rup.dict_params[job_tag] = dict_params
        dict_params["job_tag"] = job_tag
        dict_params["total_site"] = [ l, l, l, t, ]
        dict_params["load_config_params"] = { "twist_boundary_at_boundary" : [ 0.0, 0.0, 0.0, -0.5, ], }
        dict_params["fermion_params"] = mk_dict_fermion_params()
        if job_tag == "test-4nt8":
            dict_lanc_params = { 0: { 0: mk_lanc_params(job_tag, 0, 0), }, }
            dict_clanc_params = { 0: { 0: mk_clanc_params(job_tag, 0, 0), }, }
        else:
            dict_lanc_params = { 0: { 0: mk_lanc_params(job_tag, 0, 0), }, 1: { 0: mk_lanc_params(job_tag, 1, 0), }, }
            dict_clanc_params = { 0: { 0: mk_clanc_params(job_tag, 0, 0), }, 1: { 0: mk_clanc_params(job_tag, 1, 0), }, }
        dict_params["lanc_params"] = dict_lanc_params
        dict_params["clanc_params"] = dict_clanc_params
        dict_params["gf_ape_smear_coef"] = 0.5
        dict_params["gf_ape_smear_step"] = 30
        dict_params["prop_smear_coef"] = 0.9375
        dict_params["prop_smear_step"] = 10

setup_params()



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils cimport *
from . cimport everything as cc
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import cqlat as c
import qlat_utils as q
import numpy as np

### -------------------------------------------------------------------

def begin(int id_node, Coordinate size_node, int color = 0):
    cc.begin(id_node, size_node.xx, color)

def end(cc.bool is_preserving_cache = False):
    if not is_preserving_cache:
        q.clean_cache()
    cc.end(is_preserving_cache)

### -------------------------------------------------------------------

cdef class Geometry:

    def __cinit__(self):
        self.xx = cc.Geometry()
        self.cdata = <long>&(self.xx)

    def __init__(self, total_site = None, int multiplicity = 1):
        """
        if total_site is None: create geo uninitialized
        elif multiplicity is None: create geo with multiplicity = 1
        """
        self.xx.init(Coordinate(total_site).xx, multiplicity)

    def __imatmul__(self, Geometry v1):
        self.xx = v1.xx
        return self

    def copy(self, is_copying_data = True):
        x = Geometry()
        if is_copying_data:
            x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def total_site(self):
        return c.get_total_site_geo(self)

    def total_volume(self):
        return c.get_total_volume_geo(self)

    def local_site(self):
        return c.get_node_site_geo(self)

    def local_volume(self):
        return c.get_local_volume_geo(self)

    def multiplicity(self):
        return c.get_multiplicity_geo(self)

    def eo(self):
        return c.get_eo_geo(self)

    def expansion_left(self):
        return c.get_expansion_left_geo(self)

    def expansion_right(self):
        return c.get_expansion_right_geo(self)

    def id_node(self):
        return c.get_id_node_geo(self)

    def num_node(self):
        return c.get_num_node_geo(self)

    def coor_node(self):
        return c.get_coor_node_geo(self)

    def size_node(self):
        return c.get_size_node_geo(self)

    def __str__(self):
        return self.show()

    def __repr__(self):
        return self.show()

    def show(self):
        total_site = self.total_site()
        multiplicity = self.multiplicity()
        expan_left = self.expansion_left()
        expan_right = self.expansion_right()
        eo = self.eo()
        zero = [ 0, 0, 0, 0, ]
        if expan_left == zero and expan_right == zero and eo == 0:
            return f"Geometry({total_site}, {multiplicity})"
        else:
            return f"Geometry({total_site}, {multiplicity}, expan_left={expan_left}, expan_right={expan_right}, eo={eo})"

    def coordinate_g_from_l(self, xl):
        return c.coordinate_g_from_l_geo(self, xl)

    def coordinate_l_from_g(self, xg):
        return c.coordinate_l_from_g_geo(self, xg)

    def is_local(self, xl):
        return c.is_local_geo(self, xl)

    def is_local_xg(self, xg):
        """
        return a global coordinate is inside the local volume or not
        """
        return c.is_local_xg_geo(self, xg)

    def xg_list(self):
        """
        return xg for all local sites
        """
        return c.get_xg_list(self)

### -------------------------------------------------------------------

def geo_reform(Geometry geo, int multiplicity = 1, expansion_left = None, expansion_right = None):
    cdef Coordinate el, er
    if expansion_left is None:
        el = Coordinate()
    elif isinstance(expansion_left, int):
        e = expansion_left
        el = Coordinate([ e, e, e, e, ])
    elif isinstance(expansion_left, list):
        el = Coordinate(expansion_left)
    else:
        assert isinstance(expansion_left, Coordinate)
        el = <Coordinate>expansion_left
    if expansion_right is None:
        er = Coordinate()
    elif isinstance(expansion_right, int):
        e = expansion_right
        er = Coordinate([ e, e, e, e, ])
    elif isinstance(expansion_right, list):
        er = Coordinate(expansion_right)
    else:
        assert isinstance(expansion_left, Coordinate)
        er = <Coordinate>expansion_left
    cdef Geometry geo_new = Geometry()
    geo_new.xx = cc.geo_reform(geo.xx, multiplicity, el.xx, er.xx)
    return geo_new

def geo_eo(Geometry geo, int eo = 0):
    cdef Geometry geo_new = Geometry()
    geo_new.xx = cc.geo_eo(geo.xx, eo)
    return geo_new

### -------------------------------------------------------------------

field_ctypes_complex = [
        ElemTypeColorMatrix,
        ElemTypeWilsonMatrix,
        ElemTypeNonRelWilsonMatrix,
        ElemTypeIsospinMatrix,
        ElemTypeSpinMatrix,
        ElemTypeWilsonVector,
        ElemTypeComplex,
        ]

field_ctypes_double = [
        ElemTypeColorMatrix,
        ElemTypeWilsonMatrix,
        ElemTypeNonRelWilsonMatrix,
        ElemTypeIsospinMatrix,
        ElemTypeSpinMatrix,
        ElemTypeWilsonVector,
        ElemTypeComplex,
        ElemTypeDouble,
        ]

field_ctypes_float = [
        ElemTypeComplexF,
        ElemTypeFloat,
        ]

field_ctypes_long = [
        ElemTypeLong,
        ElemTypeInt64t,
        ]

### -------------------------------------------------------------------

field_type_dict = {}

selected_field_type_dict = {}

selected_points_type_dict = {}

def Field(type ctype, Geometry geo = None, int multiplicity = 0):
    assert ctype in field_type_dict
    FieldType = field_type_dict[ctype]
    field = FieldType(geo, multiplicity)
    return field

def SelectedField(type ctype, FieldSelection fsel, int multiplicity = 0):
    assert ctype in field_type_dict
    FieldType = selected_field_type_dict[ctype]
    field = FieldType(fsel, multiplicity)
    return field

def SelectedPoints(type ctype, PointsSelection psel, int multiplicity = 0):
    assert ctype in field_type_dict
    FieldType = selected_points_type_dict[ctype]
    field = FieldType(psel, multiplicity)
    return field

### -------------------------------------------------------------------

cdef class FieldBase:

    ctype = ElemType

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def total_site(self):
        return c.get_total_site_field(self)

    def multiplicity(self):
        return c.get_multiplicity_field(self)

    def sizeof_m(self):
        return c.get_sizeof_m_field(self)

    def geo(self):
        geo = Geometry((0, 0, 0, 0))
        c.set_geo_field(geo, self)
        return geo

    def mview(self):
        return c.get_mview_field(self)

    def __iadd__(self, f1):
        """
        f1 can be Field, SelectedField, SelectedPoints
        """
        if isinstance(f1, FieldBase):
            assert f1.ctype is self.ctype
            c.set_add_field(self, f1)
        else:
            from qlat.selected_field import SelectedField
            from qlat.selected_points import SelectedPoints
            if isinstance(f1, SelectedFieldBase):
                c.acc_field_sfield(self, f1)
            elif isinstance(f1, SelectedPointsBase):
                assert f1.ctype is self.ctype
                c.acc_field_spfield(self, f1)
            else:
                raise Exception(f"Field += type mismatch {type(self)} {type(f1)}")
        return self

    def __isub__(self, f1):
        """
        f1 can be Field, SelectedField, SelectedPoints
        """
        if isinstance(f1, FieldBase):
            assert f1.ctype is self.ctype
            c.set_sub_field(self, f1)
        else:
            from qlat.selected_field import SelectedField
            from qlat.selected_points import SelectedPoints
            if isinstance(f1, SelectedFieldBase):
                assert f1.ctype is self.ctype
                f1n = f1.copy()
                f1n *= -1
                c.acc_field_sfield(self, f1n)
            elif isinstance(f1, SelectedPointsBase):
                assert f1.ctype is self.ctype
                f1n = f1.copy()
                f1n *= -1
                c.acc_field_spfield(self, f1n)
            else:
                raise Exception(f"Field += type mismatch {type(self)} {type(f1)}")
            assert False
        return self

    def __imul__(self, factor):
        """
		factor can be float, complex, FieldM<Complex,1>
        """
        if isinstance(factor, float):
            c.set_mul_double_field(self, factor)
        elif isinstance(factor, complex):
            c.set_mul_complex_field(self, factor)
        elif isinstance(factor, FieldBase):
            assert factor.ctype in [ ElemTypeComplex, ElemTypeDouble, ]
            c.set_mul_cfield_field(self, factor)
        else:
            assert False
        return self

    @q.timer
    def set_unit(self, coef = 1.0):
        c.set_unit_field(self, coef)

    @q.timer
    def set_rand(self, rng, upper = 1.0, lower = 0.0):
        assert isinstance(rng, RngState)
        if self.ctype in field_ctypes_double:
            c.set_u_rand_double_field(self, rng, upper, lower)
        elif self.ctype in field_ctypes_float:
            c.set_u_rand_float_field(self, rng, upper, lower)
        else:
            assert False

    @q.timer
    def set_rand_g(self, rng, center = 0.0, sigma = 1.0):
        assert isinstance(rng, RngState)
        if self.ctype in field_ctypes_double:
            c.set_g_rand_double_field(self, rng, center, sigma)
        else:
            assert False

    def qnorm(self):
        return c.qnorm_field(self)

    def crc32(self):
        return c.crc32_field(self)

    def sparse(self, sel):
        """
        deprecated
        """
        q.displayln_info("Field.sparse: deprecated")
        from qlat.field_selection import PointsSelection, FieldSelection
        if isinstance(sel, PointsSelection):
            from qlat.selected_points import SelectedPoints, set_selected_points
            psel = sel
            sp = SelectedPoints(self.ctype, psel)
            set_selected_points(sp, self)
            return sp
        elif isinstance(sel, FieldSelection):
            from qlat.selected_field import SelectedField, set_selected_field
            fsel = sel
            sf = SelectedField(self.ctype, fsel)
            set_selected_field(sf, self)
            return sf
        else:
            raise Exception("Field.sparse")

    def save_direct(self, path, *args):
        """
        Generic save for Field object
        save Field directly (without any conversion of endianness or precision)
        possible way to call:
        f.save_direct(path)
        f.save_direct(path, new_size_node)
        f.save_direct(sfw, fn)
        """
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            if len(args) == 0:
                n_bytes = c.save_field(self, path)
            else:
                [ new_size_node, ] = args
                n_bytes = c.save_field(self, path, new_size_node)
        elif isinstance(path, ShuffledFieldsWriter):
            sfw = path
            [fn] = args
            n_bytes = sfw.write(fn, self)
        else:
            raise Exception("Field.save_direct")
        assert n_bytes != 0
        return n_bytes

    def load_direct(self, path, *args):
        """
        Generic load for Field object
        load Field directly (without any conversion of endianness or precision)
        Field geo and multiplicity will be determined during loading
        possible way to call:
        f.load_direct(path)
        f.load_direct(sfr, fn)
        """
        from qlat.fields_io import ShuffledFieldsReader
        if isinstance(path, str):
            n_bytes = c.load_field(self, path)
        elif isinstance(path, ShuffledFieldsReader):
            sfr = path
            [ fn, ] = args
            n_bytes = sfr.read(fn, self)
        else:
            raise Exception("Field.load_direct")
        assert n_bytes != 0
        return n_bytes

    def save_64(self, path, *args):
        """
        Generic save for 64-bit size element Field object
        save 64-bit Field (do conversion of endianness)
        """
        f = self.copy()
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            f.to_from_endianness("big_64")
        elif isinstance(path, ShuffledFieldsWriter):
            f.to_from_endianness("little_64")
        return f.save_direct(path, *args)

    def save_double(self, path, *args):
        """
        Generic save for double element Field object
        save double Field as double (do conversion of endianness)
        """
        return self.save_64(path, *args)

    def save_float_from_double(self, path, *args):
        """
        Generic save for double element Field object
        save double Field as float (do conversion of endianness and precision)
        """
        ff = FieldFloat()
        ff.float_from_double(self)
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            ff.to_from_endianness("big_32")
        elif isinstance(path, ShuffledFieldsWriter):
            ff.to_from_endianness("little_32")
        return ff.save_direct(path, *args)

    def load_64(self, path, *args):
        """
        Generic load for 64-bit size element Field object
        load 64-bit Field (do conversion of endianness)
        """
        ret = self.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                self.to_from_endianness("big_64")
            elif isinstance(path, ShuffledFieldsReader):
                self.to_from_endianness("little_64")
        return ret

    def load_double(self, path, *args):
        """
        Generic load for double Field object
        load double Field (do conversion of endianness)
        """
        return self.load_64(path, *args)

    def load_double_from_float(self, path, *args):
        """
        Generic load for double Field object
        load double Field from float(do conversion of endianness or precision)
        """
        ff = FieldFloat()
        ret = ff.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                ff.to_from_endianness("big_32")
            elif isinstance(path, ShuffledFieldsReader):
                ff.to_from_endianness("little_32")
            self.double_from_float(ff)
        return ret

    def float_from_double(self, FieldBase f):
        """
        self needs to be FieldFloat
        """
        assert isinstance(self, FieldFloat)
        c.convert_float_from_double_field(self, f)

    def double_from_float(self, FieldFloat ff):
        """
        self can be any FieldBase subtype but need to be actually contains double precision numbers
        """
        c.convert_double_from_float_field(self, ff)

    def to_from_endianness(self, tag):
        """
        Convert between the native endianness and the endianness specified by ``tag``
        tag can be ``"big_32", "big_64", "little_32", "little_64"``
        """
        assert isinstance(tag, str)
        c.to_from_endianness_field(self, tag)

    def as_field(self, ctype = ElemTypeComplex):
        """
		return new Field(ctype) with the same content
        """
        f = Field(ctype)
        c.assign_as_field(f, self)
        return f

    def from_field(self, f):
        """
		assign from f with the same content but possibly different type
        """
        c.assign_from_field(self, f)
        return f

    def get_elems(self, index):
        """
        index can also be xg
        get_elems is collective operation when xg is coordinate
        get_elems will be specific to a single process if xg is index
        """
        return np.array(c.get_elems_field(self, index))

    def get_elem(self, index, m = None):
        """
        index can also be xg
        """
        if m is None:
            return np.array(c.get_elem_field(self, index))
        else:
            return np.array(c.get_elem_field(self, index, m))

    def set_elems(self, index, val):
        """
        index can also be xg
        val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        """
        if isinstance(val, bytes):
            return c.set_elems_field(self, index, val)
        elif isinstance(val, np.ndarray):
            return self.set_elems(index, val.tobytes())
        else:
            assert False

    def set_elem(self, index, m, val):
        """
        index can also be xg
        val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        """
        if isinstance(val, bytes):
            return c.set_elem_field(self, index, m, val)
        elif isinstance(val, np.ndarray):
            return self.set_elem(index, m, val.tobytes())
        else:
            assert False

    def __getitem__(self, i):
        """
        i can be (index, m,) or index
        index can also be xg
        """
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], (int, list, tuple,)):
            index, m = i
            return self.get_elem(index, m)
        elif isinstance(i, (int, list, tuple)):
            index = i
            return self.get_elems(index)
        else:
            assert False
            return None

    def __setitem__(self, i, val):
        """
        i can be (index, m,) or index
        index can also be xg
        val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        """
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], (int, list, tuple,)):
            index, m = i
            self.set_elem(index, m, val)
        elif isinstance(i, (int, list, tuple)):
            index = i
            self.set_elems(index, val)
        else:
            assert False

    def xg_list(self):
        """
        return xg for all local sites
        """
        return self.geo().xg_list()

    def field_shift(self, shift):
        """
        return new shifted Field
        shift is the coordinate to shift the field
        """
        f1 = self.copy(is_copying_data = False)
        c.field_shift_field(f1, self, shift)
        return f1

    def reflect(self):
        """
        reflect the field, return None
        """
        return c.reflect_field(self)

    def glb_sum(self):
        if self.ctype in field_ctypes_double:
            return np.array(c.glb_sum_double_field(self))
        elif self.ctype in field_ctypes_long:
            return np.array(c.glb_sum_long_field(self))
        else:
            assert False
            return None

    def glb_sum_tslice(self, *, t_dir = 3):
        """
        return SelectedPoints(self.ctype, get_psel_tslice(self.total_site(), t_dir = t_dir))
        """
        from qlat.field_selection import get_psel_tslice
        from qlat.selected_points import SelectedPoints
        psel = get_psel_tslice(self.total_site(), t_dir = t_dir)
        sp = SelectedPoints(self.ctype, psel)
        if self.ctype in field_ctypes_double:
            c.glb_sum_tslice_double_field(sp, self, t_dir)
            return sp
        elif self.ctype in field_ctypes_long:
            c.glb_sum_tslice_long_field(sp, self, t_dir)
            return sp
        else:
            assert False
            return None

### -------------------------------------------------------------------

def split_fields(fs, f):
    nf = len(fs)
    assert nf >= 1
    ctype = f.ctype
    for i in range(nf):
        if not isinstance(fs[i], FieldBase):
            fs[i] = Field(ctype)
        else:
            assert fs[i].ctype is ctype
    c.split_fields_field(fs, f)

def merge_fields(f, fs):
    nf = len(fs)
    assert nf >= 1
    assert isinstance(f, FieldBase)
    assert f.ctype is fs[0].ctype
    c.merge_fields_field(f, fs)

def merge_fields_ms(f, fms):
    """
    fms = [ (f0, m0,), (f1, m1,), ... ]
    f.get_elem(x, m) = fms[m][0].get_elem(x, fms[m][1])
    """
    multiplicity = len(fms)
    assert multiplicity >= 1
    assert isinstance(f, FieldBase)
    assert f.ctype is fms[0][0].ctype
    fs, ms = zip(*fms)
    c.merge_fields_ms_field(f, fs, ms)

def mk_merged_fields_ms(fms):
    """
    fms = [ (f0, m0,), (f1, m1,), ... ]
    f.get_elem(x, m) = fms[m][0].get_elem(x, fms[m][1])
    return f
    """
    multiplicity = len(fms)
    assert multiplicity >= 1
    for m in range(multiplicity):
        assert isinstance(fms[m][0], FieldBase)
        assert isinstance(fms[m][1], int)
    ctype = fms[0][0].ctype
    for m in range(multiplicity):
        assert ctype is fms[m][0].ctype
    f = Field(ctype)
    merge_fields_ms(f, fms)
    return f

### -------------------------------------------------------------------

cdef class PointsSelection:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)
        self.geo = None

    def __init__(self, coordinate_list = None, Geometry geo = None):
        cdef long n_points
        cdef long i
        self.geo = geo
        if coordinate_list is None:
            return
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        n_points = len(coordinate_list)
        self.xx = cc.PointsSelection(n_points)
        for i in range(n_points):
            c = coordinate_list[i]
            if not isinstance(c, Coordinate):
                c = Coordinate(c)
            self.xx[i] = (<Coordinate>c).xx

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef long n_points = self.xx.size()
        cdef int ndim = 2
        cdef char* fmt = "i"
        cdef Buffer buf = Buffer(self, ndim, sizeof(cc.Int32t))
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        shape[0] = n_points
        shape[1] = 4
        buf.set_strides()
        buffer.buf = <char*>(self.xx.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL
        self.view_count += 1

    def __releasebuffer__(self, Py_buffer *buffer):
        self.view_count -= 1

    def __imatmul__(self, PointsSelection v1):
        self.geo = v1.geo
        self.xx = v1.xx
        return self

    def copy(self):
        x = PointsSelection()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_rand(self, rs, total_site, n_points):
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        c.set_rand_psel(self, rs, total_site, n_points)
        self.geo = Geometry(total_site)

    def save(self, path):
        c.save_psel(self, path)

    def load(self, path, geo):
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        c.load_psel(self, path)
        self.geo = geo

    def n_points(self):
        return c.get_n_points_psel(self)

    def to_list(self):
        return c.mk_list_psel(self)

    def from_list(self, coordinate_list, geo = None):
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        c.set_list_psel(self, coordinate_list)
        self.geo = geo
        return self

    def coordinate_from_idx(self, idx):
        return c.get_coordinate_from_idx_psel(self, idx)

### -------------------------------------------------------------------

cdef class FieldSelection:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __init__(self, total_site = None, n_per_tslice = -1, rs = None, psel = None):
        if total_site is not None:
            assert isinstance(rs, RngState)
            assert isinstance(n_per_tslice, int)
            c.set_rand_fsel(self, rs, total_site, n_per_tslice)
            if psel is not None:
                c.add_psel_fsel(self, psel)
            self.update()
            self.update(n_per_tslice)

    def __imatmul__(self, FieldSelection v1):
        self.xx = v1.xx
        return self

    def copy(self):
        x = FieldSelection()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_uniform(self, total_site, val = 0):
        """
        default (val = 0) select every sites
        val = -1 deselection everything
        """
        c.set_uniform_fsel(self, total_site, val)

    def set_rand(self, rs, total_site, n_per_tslice):
        assert isinstance(rs, RngState)
        assert isinstance(n_per_tslice, int)
        c.set_rand_fsel(self, rs, total_site, n_per_tslice)
        self.update()
        self.update(n_per_tslice)

    def add_psel(self, psel, rank_psel = 1024 * 1024 * 1024 * 1024 * 1024):
        """
        Add psel points to the selection, with the rank specified as rank_psel.
        If the point is already selected with lower rank, the rank is unchanged.
        """
        c.add_psel_fsel(self, psel, rank_psel)
        self.update()

    def update(self, n_per_tslice = -1):
        """
        if n_per_tslice < 0: only update various indices
        if n_per_tslice >= 0: only update parameters (n_per_tslice and prob)
        """
        c.update_fsel(self, n_per_tslice)

    def select_rank_range(self, rank_start = 0, rank_stop = -1):
        """
        return new fsel with selected points that
        rank_start <= rank and (rank < rank_stop or rank_stop == -1)
        Does NOT change the n_per_tslice parameter for the new fsel
        """
        fsel = FieldSelection()
        c.select_rank_range_fsel(fsel, self, rank_start, rank_stop)
        fsel.update()
        fsel.update(self.n_per_tslice())
        return fsel

    def select_t_range(self, rank_start = 0, rank_stop = -1):
        """
        return new fsel with selected points that
        t_start <= t and (t < t_stop or t_stop == -1)
        rank_start <= rank < rank_stop (rank_stop = -1 implies unlimited)
        Does NOT change the n_per_tslice parameter for the new fsel
        """
        fsel = FieldSelection()
        c.select_rank_range_fsel(fsel, self, rank_start, rank_stop)
        fsel.update()
        fsel.update(self.n_per_tslice())
        return fsel

    def to_psel(self):
        psel = PointsSelection(None, self.geo())
        c.set_psel_fsel(psel, self)
        return psel

    def to_psel_local(self):
        psel = PointsSelection(None, self.geo())
        c.set_psel_fsel_local(psel, self)
        return psel

    def save(self, path):
        return c.save_fsel(self, path)

    def load(self, path, n_per_tslice):
        return c.load_fsel(self, path, n_per_tslice)

    def geo(self):
        geo = Geometry((0, 0, 0, 0))
        c.set_geo_fsel(geo, self)
        return geo

    def total_site(self):
        return c.get_total_site_fsel(self)

    def n_elems(self):
        return c.get_n_elems_fsel(self)

    def n_per_tslice(self):
        return c.get_n_per_tslice_fsel(self)

    def prob(self):
        """
        return fsel.prob
        n_per_tslice / spatial_volume
        """
        return c.get_prob_fsel(self)

    def idx_from_coordinate(self, xg):
        return c.get_idx_from_coordinate_fsel(self, xg)

    def coordinate_from_idx(self, idx):
        return c.get_coordinate_from_idx_fsel(self, idx)

### -------------------------------------------------------------------

cdef class SelectedFieldBase:

    ctype = ElemType

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def n_elems(self):
        return c.get_n_elems_sfield(self)

    def total_site(self):
        return c.get_total_site_sfield(self)

    def multiplicity(self):
        return c.get_multiplicity_sfield(self)

    def geo(self):
        geo = Geometry((0, 0, 0, 0))
        c.set_geo_sfield(geo, self)
        return geo

    def __iadd__(self, f1):
        assert isinstance(f1, SelectedFieldBase)
        assert f1.ctype is self.ctype
        c.set_add_sfield(self, f1)
        return self

    def __isub__(self, f1):
        assert isinstance(f1, SelectedFieldBase)
        assert f1.ctype is self.ctype
        c.set_sub_sfield(self, f1)
        return self

    def __imul__(self, factor):
        assert isinstance(factor, float)
        c.set_mul_double_sfield(self, factor)
        return self

    def set_rand(self, rng, upper = 1.0, lower = 0.0):
        assert isinstance(rng, RngState)
        if self.ctype in field_ctypes_double:
            c.set_u_rand_double_sfield(self, rng, upper, lower)
        else:
            assert False

    def qnorm(self):
        return c.qnorm_sfield(self)

    def get_elems(self, idx):
        return np.array(c.get_elems_sfield(self, idx))

    def get_elem(self, idx, m = None):
        if m is None:
            return np.array(c.get_elem_sfield(self, idx))
        else:
            return np.array(c.get_elem_sfield(self, idx, m))

    def set_elems(self, idx, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elems_sfield(self, idx, val)
        elif isinstance(val, np.ndarray):
            return self.set_elems(idx, val.tobytes())
        else:
            assert False

    def set_elem(self, idx, m, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elem_sfield(self, idx, m, val)
        elif isinstance(val, np.ndarray):
            return self.set_elem(idx, m, val.tobytes())
        else:
            assert False

    def __getitem__(self, i):
        # i can be (idx, m,) or idx
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            return self.get_elem(idx, m)
        elif isinstance(i, int):
            idx = i
            return self.get_elems(idx)
        else:
            assert False
            return None

    def __setitem__(self, i, val):
        # i can be (idx, m,) or idx
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            self.set_elem(idx, m, val)
        elif isinstance(i, int):
            idx = i
            self.set_elems(idx, val)
        else:
            assert False

    def sparse(self, sel):
        """
        deprecated
        """
        q.displayln_info("SelectedField.sparse: deprecated")
        if isinstance(sel, PointsSelection):
            from qlat.selected_points import SelectedPoints, set_selected_points
            psel = sel
            sp = SelectedPoints(self.ctype, psel)
            set_selected_points(sp, self)
            return sp
        elif isinstance(sel, FieldSelection):
            from qlat.selected_field import SelectedField, set_selected_field
            fsel = sel
            sf = SelectedField(self.ctype, fsel)
            set_selected_field(sf, self)
            return sf
        else:
            raise Exception("Field.sparse")

    def save_direct(self, path, *args):
        """
        Generic save for SelectedField object
        possible way to call:
        f.save_direct(path) # has some limitations
        f.save_direct(sfw, fn)
        """
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            assert len(args) == 0
            n_bytes = c.save_sfield(self, path)
        elif isinstance(path, ShuffledFieldsWriter):
            sfw = path
            [fn] = args
            n_bytes = sfw.write(fn, self)
        else:
            raise Exception("SelectedField.save_direct")
        assert n_bytes != 0
        return n_bytes

    def load_direct(self, path, *args):
        """
        Generic load for SelectedField object
        possible way to call:
        f.load_direct(path) # has some limitations
        f.load_direct(sfr, fn)
        if self.fsel is None, self.fsel will be set during f.load_direct(sfr, fn)
        """
        from qlat.fields_io import ShuffledFieldsReader
        if isinstance(path, str):
            n_bytes = c.load_sfield(self, path)
        elif isinstance(path, ShuffledFieldsReader):
            sfr = path
            [fn] = args
            n_bytes = sfr.read(fn, self)
        else:
            raise Exception("SelectedField.load")
        assert n_bytes != 0
        return n_bytes

    def save_64(self, path, *args):
        """
        Generic save for SelectedField object with conversion
        """
        f = self.copy()
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            f.to_from_endianness("big_64")
        elif isinstance(path, ShuffledFieldsWriter):
            f.to_from_endianness("little_64")
        return f.save_direct(path, *args)

    def save_double(self, path, *args):
        """
        Generic save for SelectedField object with conversion
        """
        return self.save_64(path, *args)

    def save_float_from_double(self, path, *args):
        """
        Generic save for SelectedField object with conversion
        """
        ff = SelectedFieldFloat(self.fsel)
        ff.float_from_double(self)
        from qlat.fields_io import ShuffledFieldsWriter
        if isinstance(path, str):
            ff.to_from_endianness("big_32")
        elif isinstance(path, ShuffledFieldsWriter):
            ff.to_from_endianness("little_32")
        return ff.save_direct(path, *args)

    def load_64(self, path, *args):
        """
        Generic load for SelectedField object with conversion
        """
        ret = self.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                self.to_from_endianness("big_64")
            elif isinstance(path, ShuffledFieldsReader):
                self.to_from_endianness("little_64")
        return ret

    def load_double(self, path, *args):
        """
        Generic load for SelectedField object with conversion
        """
        return self.load_64(path, *args)

    def load_double_from_float(self, path, *args):
        """
        Generic load for SelectedField object with conversion
        """
        ff = SelectedField(ElemTypeFloat, self.fsel)
        ret = ff.load_direct(path, *args)
        if ret > 0:
            from qlat.fields_io import ShuffledFieldsReader
            if isinstance(path, str):
                ff.to_from_endianness("big_32")
            elif isinstance(path, ShuffledFieldsReader):
                ff.to_from_endianness("little_32")
            self.double_from_float(ff)
        return ret

    def float_from_double(self, SelectedFieldBase f):
        assert isinstance(self, SelectedFieldFloat)
        self.fsel = f.fsel
        c.convert_float_from_double_sfield(self, f)

    def double_from_float(self, SelectedFieldFloat ff):
        self.fsel = ff.fsel
        c.convert_double_from_float_sfield(self, ff)

    def to_from_endianness(self, tag):
        assert isinstance(tag, str)
        c.to_from_endianness_sfield(self, tag)

    def field_shift(self, shift, is_reflect = False):
        # return new shifted SelectedField
        # shift is the coordinate to shift the field
        # is_reflect determine whether to negate coordinate after shift
        f1 = self.copy(is_copying_data = False)
        f1.fsel = FieldSelection()
        c.field_shift_sfield(f1, self, shift, is_reflect)
        return f1

    def glb_sum_tslice(self, *, t_dir = 3):
        """
        return SelectedPoints(self.ctype, get_psel_tslice(self.total_site(), t_dir = t_dir))
        """
        from qlat.selected_points import SelectedPoints, get_psel_tslice
        psel = get_psel_tslice(self.total_site(), t_dir = t_dir)
        sp = SelectedPoints(self.ctype, psel)
        if self.ctype in field_ctypes_double:
            c.glb_sum_tslice_double_sfield(sp, self, t_dir)
        elif self.ctype in field_ctypes_long:
            c.glb_sum_tslice_long_sfield(sp, self, t_dir)
        else:
            assert False
        return sp

### -------------------------------------------------------------------

cdef class SelectedPointsBase:

    ctype = ElemType

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def n_points(self):
        return c.get_n_points_spfield(self)

    def multiplicity(self):
        return c.get_multiplicity_spfield(self)

    def __iadd__(self, f1):
        assert isinstance(f1, SelectedPointsBase) and f1.ctype is self.ctype
        c.set_add_spfield(self, f1)
        return self

    def __isub__(self, f1):
        assert isinstance(f1, SelectedPointsBase) and f1.ctype is self.ctype
        c.set_sub_spfield(self, f1)
        return self

    def __imul__(self, factor):
        assert isinstance(factor, float)
        c.set_mul_double_spfield(self, factor)
        return self

    def qnorm(self):
        return c.qnorm_spfield(self)

    def get_elems(self, idx):
        return np.array(c.get_elems_spfield(self, idx))

    def get_elem(self, idx, m = None):
        if m is None:
            return np.array(c.get_elem_spfield(self, idx))
        else:
            return np.array(c.get_elem_spfield(self, idx, m))

    def set_elems(self, idx, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elems_spfield(self, idx, val)
        elif isinstance(val, np.ndarray):
            return self.set_elems(idx, val.tobytes())
        else:
            assert False

    def set_elem(self, idx, m, val):
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(val, bytes):
            return c.set_elem_spfield(self, idx, m, val)
        elif isinstance(val, np.ndarray):
            return self.set_elem(idx, m, val.tobytes())
        else:
            assert False

    def __getitem__(self, i):
        # i can be (idx, m,) or idx
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            return self.get_elem(idx, m)
        elif isinstance(i, int):
            idx = i
            return self.get_elems(idx)
        else:
            assert False
            return None

    def __setitem__(self, i, val):
        # i can be (idx, m,) or idx
        # val should be np.ndarray or bytes. e.g. np.array([1, 2, 3], dtype = complex).tobytes()
        if isinstance(i, tuple) and len(i) == 2 and isinstance(i[0], int):
            idx, m = i
            self.set_elem(idx, m, val)
        elif isinstance(i, int):
            idx = i
            self.set_elems(idx, val)
        else:
            assert False

    def save(self, path):
        assert isinstance(path, str)
        return self.save_complex(path)

    def load(self, path):
        assert isinstance(path, str)
        return self.load_complex(path)

    def save_complex(self, path):
        assert isinstance(path, str)
        return c.save_complex_spfield(self, path)

    def load_complex(self, path):
        assert isinstance(path, str)
        return c.load_complex_spfield(self, path)

    def to_lat_data(self):
        assert self.ctype in field_ctypes_complex
        ld = q.LatData()
        c.lat_data_from_complex_spfield(ld, self)
        return ld

    def from_lat_data(self, ld):
        assert self.ctype in field_ctypes_complex
        assert isinstance(ld, q.LatData)
        c.complex_spfield_from_lat_data(self, ld)

    def to_numpy(self):
        return np.asarray(self).copy()

    def from_numpy(self, arr):
        """
        need to be already initialized with ctype and psel
        arr.shape[0] == n_points
        """
        n_points = self.n_points()
        assert arr.shape[0] == n_points
        np.asarray(self).ravel()[:] = arr.ravel()

### -------------------------------------------------------------------

def qremove_info(path):
    return cc.qremove_info(path)

def qremove_all_info(path):
    return cc.qremove_all_info(path)

### -------------------------------------------------------------------



from qlat_utils import *
import qlat.c as c

from qlat.field import *
from qlat.field_utils import *
from qlat.selected_field import *
from qlat.selected_points import *
from qlat.mat import *

class Prop(FieldWilsonMatrix):

    def __init__(self, geo = None):
        super().__init__(geo, 1)

    def copy(self, is_copying_data = True):
        f = Prop()
        if is_copying_data:
            f @= self
        return f

    def get_elem_wm(self, xg, m = 0):
        wm = WilsonMatrix()
        c.get_elem_wm_prop(wm, self, xg, m)
        return wm

    def sparse(self, sel):
        # deprecated
        if isinstance(sel, PointsSelection):
            psel = sel
            sp = PselProp(psel)
            set_selected_points(sp, self)
            return sp
        elif isinstance(sel, FieldSelection):
            fsel = sel
            sf = SelProp(fsel)
            set_selected_field(sf, self)
            return sf
        else:
            raise Exception("Prop.sparse")

###

class SelProp(SelectedFieldWilsonMatrix):

    def __init__(self, fsel):
        super().__init__(fsel, 1)

    def copy(self, is_copying_data = True):
        f = SelProp(self.fsel)
        if is_copying_data:
            f @= self
        return f

    def get_elem_wm(self, idx, m = 0):
        wm = WilsonMatrix()
        cdata = c.get_elem_wm_sprop(wm, self, idx, m)
        return wm

    def sparse(self, sel):
        # deprecated
        if isinstance(sel, PointsSelection):
            psel = sel
            sp = PselProp(psel)
            set_selected_points(sp, self)
            return sp
        elif isinstance(sel, FieldSelection):
            fsel = sel
            sf = SelProp(fsel)
            set_selected_field(sf, self)
            return sf
        else:
            raise Exception("SelProp.sparse")

###

class PselProp(SelectedPointsWilsonMatrix):

    def __init__(self, psel):
        super().__init__(psel, 1)

    def copy(self, is_copying_data = True):
        f = PselProp(self.psel)
        if is_copying_data:
            f @= self
        return f

    def get_elem_wm(self, idx, m = 0):
        wm = WilsonMatrix()
        c.get_elem_wm_psprop(wm, self, idx, m)
        return wm

###

def set_point_src(prop_src, geo, xg, value = 1.0):
    c.set_point_src_prop(prop_src, geo, xg, value)

def set_wall_src(prop_src, geo, tslice, lmom = None):
    if lmom is None:
        lmom = [ 0.0, 0.0, 0.0, 0.0, ]
    c.set_wall_src_prop(prop_src, geo, tslice, lmom)

def mk_point_src(geo, xg, value = 1.0):
    prop_src = Prop()
    set_point_src(prop_src, geo, xg, value)
    return prop_src

def mk_wall_src(geo, tslice, lmom = None):
    if lmom is None:
        lmom = [ 0.0, 0.0, 0.0, 0.0, ]
    prop_src = Prop()
    set_wall_src(prop_src, geo, tslice, lmom)
    return prop_src

@timer
def mk_rand_u1_src(sel, rs):
    # return (prop_src, fu1,) where prop_src = Prop() and fu1 = Field(c.ElemTypeComplex)
    # fu1 stores the random u1 numbers (fu1.multiplicity() == 1)
    # sel can be psel or fsel
    prop_src = Prop()
    fu1 = Field(c.ElemTypeComplex)
    if isinstance(sel, FieldSelection):
        fsel = sel
        c.set_rand_u1_src_fsel(prop_src, fu1, fsel, rs)
    elif isinstance(sel, PointsSelection):
        psel = sel
        geo = psel.geo
        assert isinstance(geo, Geometry)
        c.set_rand_u1_src_psel(prop_src, fu1, psel, geo, rs)
    else:
        raise Exception(f"mk_rand_u1_src {type(sel)}")
    return (prop_src, fu1,)

@timer
def get_rand_u1_sol(prop_sol, fu1, sel):
    assert isinstance(prop_sol, Prop)
    assert isinstance(fu1, FieldBase) and fu1.ctype == c.ElemTypeComplex
    if isinstance(sel, FieldSelection):
        fsel = sel
        s_prop = SelProp(fsel)
        c.set_rand_u1_sol_fsel(s_prop, prop_sol, fu1, fsel)
        return s_prop
    elif isinstance(sel, PointsSelection):
        psel = sel
        sp_prop = PselProp(psel)
        c.set_rand_u1_sol_psel(sp_prop, prop_sol, fu1, psel)
        return sp_prop
    else:
        raise Exception(f"get_rand_u1_sol {type(sel)}")

@timer_verbose
def mk_rand_u1_prop(inv, sel, rs):
    # interface function
    # return s_prop
    # sel can be psel or fsel
    prop_src, fu1 = mk_rand_u1_src(sel, rs)
    prop_sol = inv * prop_src
    return get_rand_u1_sol(prop_sol, fu1, sel)

@timer
def free_invert(prop_src, mass, m5 = 1.0, momtwist = None):
    assert isinstance(prop_src, Prop)
    if momtwist is None:
        momtwist = [ 0.0, 0.0, 0.0, 0.0, ]
    prop_sol = Prop()
    c.free_invert_prop(prop_sol, prop_src, mass, m5, momtwist)
    return prop_sol

def convert_mspincolor_from_wm(prop_wm):
    prop_msc = prop_wm.copy(False)
    if isinstance(prop_wm, Prop):
        c.convert_mspincolor_from_wm_prop(prop_msc, prop_wm)
    elif isinstance(prop_wm, SelProp):
        c.convert_mspincolor_from_wm_s_prop(prop_msc, prop_wm)
    elif isinstance(prop_wm, PselProp):
        c.convert_mspincolor_from_wm_sp_prop(prop_msc, prop_wm)
    else:
        raise Exception("prop type match failed")
    return prop_msc

def convert_wm_from_mspincolor(prop_msc):
    prop_wm = prop_msc.copy(False)
    if isinstance(prop_msc, Prop):
        c.convert_wm_from_mspincolor_prop(prop_wm, prop_msc)
    elif isinstance(prop_msc, SelProp):
        c.convert_wm_from_mspincolor_s_prop(prop_wm, prop_msc)
    elif isinstance(prop_msc, PselProp):
        c.convert_wm_from_mspincolor_sp_prop(prop_wm, prop_msc)
    else:
        raise Exception("prop type match failed")
    return prop_wm

@timer
def flip_tpbc_with_tslice(prop, tslice_flip_tpbc):
    if isinstance(prop, SelProp):
        c.flip_tpbc_with_tslice_s_prop(prop, tslice_flip_tpbc)
    elif isinstance(prop, PselProp):
        c.flip_tpbc_with_tslice_sp_prop(prop, tslice_flip_tpbc)
    else:
        print(type(prop))
        assert False

@timer
def free_scalar_invert_mom_cfield(f, mass):
    assert isinstance(f, FieldBase)
    assert f.ctype == c.ElemTypeComplex
    c.free_scalar_invert_mom_cfield(f, mass)

@timer
def free_scalar_invert_cfield(src, mass):
    fft_f = mk_fft(True, is_normalizing = True)
    fft_b = mk_fft(False, is_normalizing = True)
    f = fft_f * src
    free_scalar_invert_mom_cfield(f, mass)
    sol = fft_b * f
    return sol

class FermionField4d(FieldWilsonVector):

    def __init__(self, geo = None):
        super().__init__(geo, 1)

    def copy(self, is_copying_data = True):
        f = FermionField4d()
        if is_copying_data:
            f @= self
        return f

###



from qlat_utils import *
import qlat.c as c

from qlat.field import *
from qlat.propagator import *

@timer
def contract_chvp_16(prop1, prop2):
    chvp_16 = Field(c.ElemTypeComplex)
    c.contract_chvp_16_field(chvp_16, prop1, prop2)
    return chvp_16



#    Qlattice (https://github.com/jinluchang/qlattice)
#
#    Copyright (C) 2023
#
#    Author: Luchang Jin (ljin.luchang@gmail.com)
#    Author: Masaaki Tomii
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import numpy as np

from qlat import \
        timer, \
        timer_flops

from qlat_utils.ama import \
        ama_list, \
        ama_apply1, \
        ama_counts, \
        ama_extract

from auto_contractor.distillation_mat_op import \
        load_prop, \
        get_gamma_matrix, \
        wilson_matrix_g5_herm, \
        mat_tr_sm, \
        mat_tr_cm, \
        mat_tr_wm, \
        mat_tr_wm_wm, \
        mat_tr_wm_sm, \
        mat_tr_sm_wm, \
        mat_tr_sm_sm, \
        mat_tr_wm_cm, \
        mat_tr_cm_wm, \
        mat_tr_cm_cm, \
        mat_mul_wm_wm, \
        mat_mul_wm_sm, \
        mat_mul_sm_wm, \
        mat_mul_sm_sm, \
        mat_mul_wm_cm, \
        mat_mul_cm_wm, \
        mat_mul_cm_cm

from . import auto_fac_funcs as aff



# Author: Luchang Jin 2022

import qlat as q
import sys

if len(sys.argv) < 2:
    q.displayln_info("Usage: fields-list path1 path2 ...")
    exit()

q.begin_with_mpi()

for path in sys.argv[1:]:
    tags = q.list_fields(path)
    for tag in tags:
        q.displayln_info(tag)

q.timer_display()

q.end_with_mpi()

exit()



import qlat.c as c

from qlat.hmc import *
from qlat.utils_io import *

def get_gm_force_magnitudes(gm_force, n_elems):
    return c.get_gm_force_magnitudes(gm_force, n_elems)

def display_gm_force_magnitudes(gm_force, n_elems):
    c.display_gm_force_magnitudes(gm_force, n_elems)

def save_gm_force_magnitudes_list(fn):
    mk_file_dirs_info(fn)
    c.save_gm_force_magnitudes_list(fn)

def display_gauge_field_info_table_with_wilson_flow(
        fn_gf_info, fn_wilson_flow_energy, gf, flow_time, flow_steps, steps, c1 = 0.0):
    c.display_gauge_field_info_table_with_wilson_flow(
            fn_gf_info, fn_wilson_flow_energy, gf, flow_time, flow_steps, steps, c1)

